@@ -3,12 +3,12 @@
  * tupdesc.c
  *	  POSTGRES tuple descriptor support code
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/access/common/tupdesc.c,v 1.76 2001/10/25 05:49:20 momjian Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/access/common/tupdesc.c,v 1.100 2003/09/25 06:57:56 petere Exp $
  *
  * NOTES
  *	  some of the executor utility code such as "ExecTypeFromTL" should be
@@ -19,10 +19,14 @@
 
 #include "postgres.h"
 
+#include "funcapi.h"
+#include "access/heapam.h"
+#include "catalog/namespace.h"
 #include "catalog/pg_type.h"
 #include "nodes/parsenodes.h"
 #include "parser/parse_type.h"
 #include "utils/builtins.h"
+#include "utils/lsyscache.h"
 #include "utils/syscache.h"
 
 
@@ -33,28 +37,34 @@
  * ----------------------------------------------------------------
  */
 TupleDesc
-CreateTemplateTupleDesc(int natts)
+CreateTemplateTupleDesc(int natts, bool hasoid)
 {
-	uint32		size;
 	TupleDesc	desc;
 
 	/*
 	 * sanity checks
 	 */
-	AssertArg(natts >= 1);
+	AssertArg(natts >= 0);
 
 	/*
 	 * allocate enough memory for the tuple descriptor and zero it as
 	 * TupleDescInitEntry assumes that the descriptor is filled with NULL
 	 * pointers.
 	 */
-	size = natts * sizeof(Form_pg_attribute);
 	desc = (TupleDesc) palloc(sizeof(struct tupleDesc));
-	desc->attrs = (Form_pg_attribute *) palloc(size);
-	desc->constr = NULL;
-	MemSet(desc->attrs, 0, size);
 
 	desc->natts = natts;
+	desc->tdhasoid = hasoid;
+
+	if (natts > 0)
+	{
+		uint32		size = natts * sizeof(Form_pg_attribute);
+
+		desc->attrs = (Form_pg_attribute *) palloc0(size);
+	}
+	else
+		desc->attrs = NULL;
+	desc->constr = NULL;
 
 	return desc;
 }
@@ -62,23 +72,25 @@ CreateTemplateTupleDesc(int natts)
 /* ----------------------------------------------------------------
  *		CreateTupleDesc
  *
- *		This function allocates a new TupleDesc from Form_pg_attribute array
+ *		This function allocates a new TupleDesc pointing to a given
+ *		Form_pg_attribute array
  * ----------------------------------------------------------------
  */
 TupleDesc
-CreateTupleDesc(int natts, Form_pg_attribute *attrs)
+CreateTupleDesc(int natts, bool hasoid, Form_pg_attribute *attrs)
 {
 	TupleDesc	desc;
 
 	/*
 	 * sanity checks
 	 */
-	AssertArg(natts >= 1);
+	AssertArg(natts >= 0);
 
 	desc = (TupleDesc) palloc(sizeof(struct tupleDesc));
 	desc->attrs = attrs;
 	desc->natts = natts;
 	desc->constr = NULL;
+	desc->tdhasoid = hasoid;
 
 	return desc;
 }
@@ -101,18 +113,22 @@ CreateTupleDescCopy(TupleDesc tupdesc)
 
 	desc = (TupleDesc) palloc(sizeof(struct tupleDesc));
 	desc->natts = tupdesc->natts;
-	size = desc->natts * sizeof(Form_pg_attribute);
-	desc->attrs = (Form_pg_attribute *) palloc(size);
-	for (i = 0; i < desc->natts; i++)
+	if (desc->natts > 0)
 	{
-		desc->attrs[i] = (Form_pg_attribute) palloc(ATTRIBUTE_TUPLE_SIZE);
-		memmove(desc->attrs[i],
-				tupdesc->attrs[i],
-				ATTRIBUTE_TUPLE_SIZE);
-		desc->attrs[i]->attnotnull = false;
-		desc->attrs[i]->atthasdef = false;
+		size = desc->natts * sizeof(Form_pg_attribute);
+		desc->attrs = (Form_pg_attribute *) palloc(size);
+		for (i = 0; i < desc->natts; i++)
+		{
+			desc->attrs[i] = (Form_pg_attribute) palloc(ATTRIBUTE_TUPLE_SIZE);
+			memcpy(desc->attrs[i], tupdesc->attrs[i], ATTRIBUTE_TUPLE_SIZE);
+			desc->attrs[i]->attnotnull = false;
+			desc->attrs[i]->atthasdef = false;
+		}
 	}
+	else
+		desc->attrs = NULL;
 	desc->constr = NULL;
+	desc->tdhasoid = tupdesc->tdhasoid;
 
 	return desc;
 }
@@ -122,7 +138,6 @@ CreateTupleDescCopy(TupleDesc tupdesc)
  *
  *		This function creates a new TupleDesc by copying from an existing
  *		TupleDesc (with Constraints)
- *
  * ----------------------------------------------------------------
  */
 TupleDesc
@@ -135,15 +150,18 @@ CreateTupleDescCopyConstr(TupleDesc tupdesc)
 
 	desc = (TupleDesc) palloc(sizeof(struct tupleDesc));
 	desc->natts = tupdesc->natts;
-	size = desc->natts * sizeof(Form_pg_attribute);
-	desc->attrs = (Form_pg_attribute *) palloc(size);
-	for (i = 0; i < desc->natts; i++)
+	if (desc->natts > 0)
 	{
-		desc->attrs[i] = (Form_pg_attribute) palloc(ATTRIBUTE_TUPLE_SIZE);
-		memmove(desc->attrs[i],
-				tupdesc->attrs[i],
-				ATTRIBUTE_TUPLE_SIZE);
+		size = desc->natts * sizeof(Form_pg_attribute);
+		desc->attrs = (Form_pg_attribute *) palloc(size);
+		for (i = 0; i < desc->natts; i++)
+		{
+			desc->attrs[i] = (Form_pg_attribute) palloc(ATTRIBUTE_TUPLE_SIZE);
+			memcpy(desc->attrs[i], tupdesc->attrs[i], ATTRIBUTE_TUPLE_SIZE);
+		}
 	}
+	else
+		desc->attrs = NULL;
 	if (constr)
 	{
 		TupleConstr *cpy = (TupleConstr *) palloc(sizeof(TupleConstr));
@@ -179,6 +197,7 @@ CreateTupleDescCopyConstr(TupleDesc tupdesc)
 	else
 		desc->constr = NULL;
 
+	desc->tdhasoid = tupdesc->tdhasoid;
 	return desc;
 }
 
@@ -189,7 +208,8 @@ FreeTupleDesc(TupleDesc tupdesc)
 
 	for (i = 0; i < tupdesc->natts; i++)
 		pfree(tupdesc->attrs[i]);
-	pfree(tupdesc->attrs);
+	if (tupdesc->attrs)
+		pfree(tupdesc->attrs);
 	if (tupdesc->constr)
 	{
 		if (tupdesc->constr->num_defval > 0)
@@ -220,9 +240,11 @@ FreeTupleDesc(TupleDesc tupdesc)
 	}
 
 	pfree(tupdesc);
-
 }
 
+/*
+ * Compare two TupleDesc structures for logical equality
+ */
 bool
 equalTupleDescs(TupleDesc tupdesc1, TupleDesc tupdesc2)
 {
@@ -232,6 +254,8 @@ equalTupleDescs(TupleDesc tupdesc1, TupleDesc tupdesc2)
 
 	if (tupdesc1->natts != tupdesc2->natts)
 		return false;
+	if (tupdesc1->tdhasoid != tupdesc2->tdhasoid)
+		return false;
 	for (i = 0; i < tupdesc1->natts; i++)
 	{
 		Form_pg_attribute attr1 = tupdesc1->attrs[i];
@@ -254,6 +278,14 @@ equalTupleDescs(TupleDesc tupdesc1, TupleDesc tupdesc2)
 			return false;
 		if (attr1->attnotnull != attr2->attnotnull)
 			return false;
+		if (attr1->atthasdef != attr2->atthasdef)
+			return false;
+		if (attr1->attisdropped != attr2->attisdropped)
+			return false;
+		if (attr1->attislocal != attr2->attislocal)
+			return false;
+		if (attr1->attinhcount != attr2->attinhcount)
+			return false;
 	}
 	if (tupdesc1->constr != NULL)
 	{
@@ -322,10 +354,10 @@ equalTupleDescs(TupleDesc tupdesc1, TupleDesc tupdesc2)
  *		a preallocated tuple descriptor.
  * ----------------------------------------------------------------
  */
-bool
+void
 TupleDescInitEntry(TupleDesc desc,
 				   AttrNumber attributeNumber,
-				   char *attributeName,
+				   const char *attributeName,
 				   Oid oidtypeid,
 				   int32 typmod,
 				   int attdim,
@@ -340,13 +372,7 @@ TupleDescInitEntry(TupleDesc desc,
 	 */
 	AssertArg(PointerIsValid(desc));
 	AssertArg(attributeNumber >= 1);
-
-	/*
-	 * attributeName's are sometimes NULL, from resdom's.  I don't know
-	 * why that is, though -- Jolly
-	 */
-/*	  AssertArg(NameIsValid(attributeName));*/
-
+	AssertArg(attributeNumber <= desc->natts);
 	AssertArg(!PointerIsValid(desc->attrs[attributeNumber - 1]));
 
 	/*
@@ -361,12 +387,17 @@ TupleDescInitEntry(TupleDesc desc,
 	 */
 	att->attrelid = 0;			/* dummy value */
 
+	/*
+	 * Note: attributeName can be NULL, because the planner doesn't always
+	 * fill in valid resname values in targetlists, particularly for resjunk
+	 * attributes.
+	 */
 	if (attributeName != NULL)
 		namestrcpy(&(att->attname), attributeName);
 	else
 		MemSet(NameStr(att->attname), 0, NAMEDATALEN);
 
-	att->attstattarget = 0;
+	att->attstattarget = -1;
 	att->attcacheoff = -1;
 	att->atttypmod = typmod;
 
@@ -376,40 +407,15 @@ TupleDescInitEntry(TupleDesc desc,
 
 	att->attnotnull = false;
 	att->atthasdef = false;
+	att->attisdropped = false;
+	att->attislocal = true;
+	att->attinhcount = 0;
 
-	/* ----------------
-	 *	search the system cache for the type tuple of the attribute
-	 *	we are creating so that we can get the typeid and some other
-	 *	stuff.
-	 *
-	 *	Note: in the special case of
-	 *
-	 *		create EMP (name = text, manager = EMP)
-	 *
-	 *	RelationNameCreateHeapRelation() calls BuildDesc() which
-	 *	calls this routine and since EMP does not exist yet, the
-	 *	system cache lookup below fails.  That's fine, but rather
-	 *	then doing a elog(ERROR) we just leave that information
-	 *	uninitialized, return false, then fix things up later.
-	 *	-cim 6/14/90
-	 * ----------------
-	 */
 	tuple = SearchSysCache(TYPEOID,
 						   ObjectIdGetDatum(oidtypeid),
 						   0, 0, 0);
 	if (!HeapTupleIsValid(tuple))
-	{
-		/*
-		 * here type info does not exist yet so we just fill the attribute
-		 * with dummy information and return false.
-		 */
-		att->atttypid = InvalidOid;
-		att->attlen = (int16) 0;
-		att->attbyval = (bool) 0;
-		att->attalign = 'i';
-		att->attstorage = 'p';
-		return false;
-	}
+		elog(ERROR, "cache lookup failed for type %u", oidtypeid);
 
 	/*
 	 * type info exists so we initialize our attribute information from
@@ -417,7 +423,7 @@ TupleDescInitEntry(TupleDesc desc,
 	 */
 	typeForm = (Form_pg_type) GETSTRUCT(tuple);
 
-	att->atttypid = tuple->t_data->t_oid;
+	att->atttypid = HeapTupleGetOid(tuple);
 
 	/*
 	 * There are a couple of cases where we must override the information
@@ -432,9 +438,8 @@ TupleDescInitEntry(TupleDesc desc,
 	 *
 	 * (Why not just make the atttypid point to the OID type, instead of the
 	 * type the query returns?	Because the executor uses the atttypid to
-	 * tell the front end what type will be returned (in BeginCommand),
-	 * and in the end the type returned will be the result of the query,
-	 * not an OID.)
+	 * tell the front end what type will be returned, and in the end the
+	 * type returned will be the result of the query, not an OID.)
 	 *
 	 * (Why not wait until the return type of the set is known (i.e., the
 	 * recursive call to the executor to execute the set has returned)
@@ -477,56 +482,19 @@ TupleDescInitEntry(TupleDesc desc,
 	}
 
 	ReleaseSysCache(tuple);
-
-	return true;
 }
 
 
-/* ----------------------------------------------------------------
- *		TupleDescMakeSelfReference
+/*
+ * BuildDescForRelation
  *
- *		This function initializes a "self-referential" attribute like
- *		manager in "create EMP (name=text, manager = EMP)".
- *		It calls TypeShellMake() which inserts a "shell" type
- *		tuple into pg_type.  A self-reference is one kind of set, so
- *		its size and byval are the same as for a set.  See the comments
- *		above in TupleDescInitEntry.
- * ----------------------------------------------------------------
- */
-static void
-TupleDescMakeSelfReference(TupleDesc desc,
-						   AttrNumber attnum,
-						   char *relname)
-{
-	Form_pg_attribute att;
-
-	att = desc->attrs[attnum - 1];
-	att->atttypid = TypeShellMake(relname);
-	att->attlen = sizeof(Oid);
-	att->attbyval = true;
-	att->attalign = 'i';
-	att->attstorage = 'p';
-	att->attndims = 0;
-}
-
-/* ----------------------------------------------------------------
- *		BuildDescForRelation
+ * Given a relation schema (list of ColumnDef nodes), build a TupleDesc.
  *
- *		This is a general purpose function identical to BuildDesc
- *		but is used by the DefineRelation() code to catch the
- *		special case where you
- *
- *				create FOO ( ..., x = FOO )
- *
- *		here, the initial type lookup for "x = FOO" will fail
- *		because FOO isn't in the catalogs yet.  But since we
- *		are creating FOO, instead of doing an elog() we add
- *		a shell type tuple to pg_type and fix things later
- *		in amcreate().
- * ----------------------------------------------------------------
+ * Note: the default assumption is no OIDs; caller may modify the returned
+ * TupleDesc if it wants OIDs.
  */
 TupleDesc
-BuildDescForRelation(List *schema, char *relname)
+BuildDescForRelation(List *schema)
 {
 	int			natts;
 	AttrNumber	attnum;
@@ -535,7 +503,6 @@ BuildDescForRelation(List *schema, char *relname)
 	AttrDefault *attrdef = NULL;
 	TupleConstr *constr = (TupleConstr *) palloc(sizeof(TupleConstr));
 	char	   *attname;
-	char		typename[NAMEDATALEN];
 	int32		atttypmod;
 	int			attdim;
 	int			ndef = 0;
@@ -545,7 +512,7 @@ BuildDescForRelation(List *schema, char *relname)
 	 * allocate a new tuple descriptor
 	 */
 	natts = length(schema);
-	desc = CreateTemplateTupleDesc(natts);
+	desc = CreateTemplateTupleDesc(natts, false);
 	constr->has_not_null = false;
 
 	attnum = 0;
@@ -553,7 +520,6 @@ BuildDescForRelation(List *schema, char *relname)
 	foreach(p, schema)
 	{
 		ColumnDef  *entry = lfirst(p);
-		List	   *arry;
 
 		/*
 		 * for each entry in the list, get the name and type information
@@ -563,41 +529,15 @@ BuildDescForRelation(List *schema, char *relname)
 		attnum++;
 
 		attname = entry->colname;
-		arry = entry->typename->arrayBounds;
 		attisset = entry->typename->setof;
 		atttypmod = entry->typename->typmod;
+		attdim = length(entry->typename->arrayBounds);
 
-		if (arry != NIL)
-		{
-			/* array of XXX is _XXX */
-			snprintf(typename, NAMEDATALEN,
-					 "_%.*s", NAMEDATALEN - 2, entry->typename->name);
-			attdim = length(arry);
-		}
-		else
-		{
-			StrNCpy(typename, entry->typename->name, NAMEDATALEN);
-			attdim = 0;
-		}
+		TupleDescInitEntry(desc, attnum, attname,
+						   typenameTypeId(entry->typename),
+						   atttypmod, attdim, attisset);
 
-		if (!TupleDescInitEntry(desc, attnum, attname,
-								typenameTypeId(typename),
-								atttypmod, attdim, attisset))
-		{
-			/*
-			 * if TupleDescInitEntry() fails, it means there is no type in
-			 * the system catalogs.  So now we check if the type name
-			 * equals the relation name.  If so we have a self reference,
-			 * otherwise it's an error.
-			 */
-			if (strcmp(typename, relname) == 0)
-				TupleDescMakeSelfReference(desc, attnum, relname);
-			else
-				elog(ERROR, "DefineRelation: no such type %s",
-					 typename);
-		}
-
-		/* This is for constraints */
+		/* Fill in additional stuff not handled by TupleDescInitEntry */
 		if (entry->is_not_null)
 			constr->has_not_null = true;
 		desc->attrs[attnum - 1]->attnotnull = entry->is_not_null;
@@ -616,7 +556,10 @@ BuildDescForRelation(List *schema, char *relname)
 			desc->attrs[attnum - 1]->atthasdef = true;
 		}
 
+		desc->attrs[attnum - 1]->attislocal = entry->is_local;
+		desc->attrs[attnum - 1]->attinhcount = entry->inhcount;
 	}
+
 	if (constr->has_not_null || ndef > 0)
 	{
 		desc->constr = constr;
@@ -645,3 +588,125 @@ BuildDescForRelation(List *schema, char *relname)
 	}
 	return desc;
 }
+
+
+/*
+ * RelationNameGetTupleDesc
+ *
+ * Given a (possibly qualified) relation name, build a TupleDesc.
+ */
+TupleDesc
+RelationNameGetTupleDesc(const char *relname)
+{
+	RangeVar   *relvar;
+	Relation	rel;
+	TupleDesc	tupdesc;
+	List	   *relname_list;
+
+	/* Open relation and get the tuple description */
+	relname_list = stringToQualifiedNameList(relname, "RelationNameGetTupleDesc");
+	relvar = makeRangeVarFromNameList(relname_list);
+	rel = relation_openrv(relvar, AccessShareLock);
+	tupdesc = CreateTupleDescCopy(RelationGetDescr(rel));
+	relation_close(rel, AccessShareLock);
+
+	return tupdesc;
+}
+
+/*
+ * TypeGetTupleDesc
+ *
+ * Given a type Oid, build a TupleDesc.
+ *
+ * If the type is composite, *and* a colaliases List is provided, *and*
+ * the List is of natts length, use the aliases instead of the relation
+ * attnames.
+ *
+ * If the type is a base type, a single item alias List is required.
+ */
+TupleDesc
+TypeGetTupleDesc(Oid typeoid, List *colaliases)
+{
+	char		functyptype = get_typtype(typeoid);
+	TupleDesc	tupdesc = NULL;
+
+	/*
+	 * Build a suitable tupledesc representing the output rows
+	 */
+	if (functyptype == 'c')
+	{
+		/* Composite data type, i.e. a table's row type */
+		Oid			relid = typeidTypeRelid(typeoid);
+		Relation	rel;
+		int			natts;
+
+		if (!OidIsValid(relid))
+			elog(ERROR, "invalid typrelid for complex type %u", typeoid);
+
+		rel = relation_open(relid, AccessShareLock);
+		tupdesc = CreateTupleDescCopy(RelationGetDescr(rel));
+		natts = tupdesc->natts;
+		relation_close(rel, AccessShareLock);
+		/* XXX should we hold the lock to ensure table doesn't change? */
+
+		if (colaliases != NIL)
+		{
+			int			varattno;
+
+			/* does the list length match the number of attributes? */
+			if (length(colaliases) != natts)
+				ereport(ERROR,
+						(errcode(ERRCODE_DATATYPE_MISMATCH),
+						 errmsg("number of aliases does not match number of columns")));
+
+			/* OK, use the aliases instead */
+			for (varattno = 0; varattno < natts; varattno++)
+			{
+				char	   *label = strVal(nth(varattno, colaliases));
+
+				if (label != NULL)
+					namestrcpy(&(tupdesc->attrs[varattno]->attname), label);
+			}
+		}
+	}
+	else if (functyptype == 'b' || functyptype == 'd')
+	{
+		/* Must be a base data type, i.e. scalar */
+		char	   *attname;
+
+		/* the alias list is required for base types */
+		if (colaliases == NIL)
+			ereport(ERROR,
+					(errcode(ERRCODE_DATATYPE_MISMATCH),
+					 errmsg("no column alias was provided")));
+
+		/* the alias list length must be 1 */
+		if (length(colaliases) != 1)
+			ereport(ERROR,
+					(errcode(ERRCODE_DATATYPE_MISMATCH),
+					 errmsg("number of aliases does not match number of columns")));
+
+		/* OK, get the column alias */
+		attname = strVal(lfirst(colaliases));
+
+		tupdesc = CreateTemplateTupleDesc(1, false);
+		TupleDescInitEntry(tupdesc,
+						   (AttrNumber) 1,
+						   attname,
+						   typeoid,
+						   -1,
+						   0,
+						   false);
+	}
+	else if (functyptype == 'p' && typeoid == RECORDOID)
+		ereport(ERROR,
+				(errcode(ERRCODE_DATATYPE_MISMATCH),
+				 errmsg("could not determine row description for function returning record")));
+	else
+	{
+		/* crummy error message, but parser should have caught this */
+		elog(ERROR, "function in FROM has unsupported return type");
+	}
+
+	return tupdesc;
+}
